table.panel-df {
  color: var(--neutral-foreground-rest);
}
.panel-df tbody tr:nth-child(odd) {
  background: var(--neutral-fill-active);
}
.panel-df thead {
  border-bottom: 1px solid var(--neutral-divider-rest);
}
.panel-df tr:hover {
  background: var(--accent-fill-hover) !important;
  color: var(--accent-foreground-cut-rest);
  cursor: pointer;
}
